using System;
using System.Collections.Generic;
using System.Data;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Globalization;
using System.IO;
using System.Text;
using System.Windows.Forms;

namespace WritingGPSApplications.Part3.Example
{
    public partial class Form1 : Form
    {
        // This array stores the latitude and longitude of Nebraska in the United States
        private PointF[] geographicCoordinates;
        // This map projection flattens 3D coordinates to 2D coordinates.
        private MapProjection plateCaree = new MapProjection();
        // This array stores the same coordinates but as projected coordinates
        private PointF[] projectedCoordinates;
        // This rectangle represents the "viewport", the portion of the coordinates we want to see.
        // The viewport defaults to a rectangle which fits the entire state of Nebraska.
        private RectangleF viewport = RectangleF.FromLTRB(
                        -11583105.0f,
                        -4787050.0f,
                        -10609638.0f,
                        -4452767.5f);
        // Use a matrix for translation and scaling
        Matrix transform = new Matrix();
        // This brush and pen are used to paint Nebraska
        private SolidBrush brush = new SolidBrush(Color.Green);
        private Pen pen = new Pen(Color.Black, 0.001f);

        public Form1()
        {
            InitializeComponent();

            // Redraw the control when it's resized
            SetStyle(ControlStyles.ResizeRedraw, true);

            #region Load Geographic Data Points

            /* This example displays the state of Nebraska on-screen.  Geographic coordinates
             * which form Nebraska are stored in a text file, Nebraska.txt.
             */
            LoadData();

            #endregion

            #region Project all Geographic Points

            /* For this example, we will "flatten" the state of Nebraska using the Plate Caree
             * map projection.  Once this is done, we can draw these coordinates onto the form.
             */

            // Next, make an array to store our projected coordinates.
            projectedCoordinates = new PointF[geographicCoordinates.Length];

            // For each geographic coordinate, project it
            for (int index = 0; index < projectedCoordinates.Length; index++)
            {
                // Convert the geographic coordinate to a projected coordinate
                projectedCoordinates[index] = plateCaree.Project(geographicCoordinates[index]);
            }

            #endregion
        }

        protected override void OnPaint(PaintEventArgs e)
        {
            e.Graphics.Clear(Color.White);

            /* In this example, the entire form surface is painted.  Since coordinates have been
             * projected, the next task is to scale and translate our "viewport" to the form's own dimensions
             */

            #region Calculate the Viewport's Aspect Ratio

            /* If the shape of the Form is different from the shape of the viewport, the reulting 
             * shape will become distorted.  To prevent this, we can make the aspect ratio of the
             * viewport match the form's aspect ratio.
             */

            // Calculate the aspect ratio of the Form
            float pixelAspectRatio = (float)this.Width / this.Height;

            // Calculate the aspect ratio of the viewport
            float projectedAspectRatio = viewport.Width / viewport.Height;

            // Make a copy of the viewport that we can change
            RectangleF adjustedViewport = viewport;

            // Is the Form's aspect ratio larger than the viewport's ratio?
            if (pixelAspectRatio > projectedAspectRatio)
            {
                // Yes.  Increase the width of the viewport
                adjustedViewport.Inflate(
                    (pixelAspectRatio * adjustedViewport.Height - adjustedViewport.Width) / 2,
                    0);
            }
            // Is the viewport's aspect ratio larger than the form's ratio?
            else if (pixelAspectRatio < projectedAspectRatio)
            {
                // Yes.  Increase the height of the viewport
                adjustedViewport.Inflate(
                    0,
                    (adjustedViewport.Width / pixelAspectRatio - adjustedViewport.Height) / 2);
            }

            #endregion

            #region Translate and Scale the Viewport

            /*
             * Projected coordinates have the same axis direction as geogrphic coordinates, 
             * North(Up) and East(Right) are positive. In pixel coordinate, the Y axis is reversed.
             * We account for this by flipping and reversing the Y values.
             */

            // Reset the transform
            transform.Reset();

            // First, translate all projected points so that they match up with pixel 0,0
            transform.Translate(-adjustedViewport.X, adjustedViewport.Y, MatrixOrder.Append);

            // Next, scale all points so that the viewport fits inside the form.
            transform.Scale(this.Width / adjustedViewport.Width,
                           this.Height / -adjustedViewport.Height, MatrixOrder.Append);

            // Apply this transform to all graphics operations
            e.Graphics.Transform = transform;

            #endregion

            #region Draw a Geographic Shape

            // Now draw nebraska using a green interior and black outline
            e.Graphics.FillPolygon(brush, projectedCoordinates);
            e.Graphics.DrawPolygon(pen, projectedCoordinates);

            #endregion
        }

        public void ZoomIn()
        {
            // First, get the width of the viewport, then calculate 10% of it
            float zoomWidthAmount = -viewport.Width * 0.10f;
            float zoomHeightAmount = -viewport.Height * 0.10f;

            // Next, apply the amounts to shrink (yes, shrink) the viewport to zoom in
            viewport.Inflate(zoomWidthAmount, zoomHeightAmount);

            // And repaint
            Invalidate();
        }

        public void ZoomOut()
        {
            // First, get the width of the viewport, then calculate 10% of it
            float zoomWidthAmount = viewport.Width * 0.10f;
            float zoomHeightAmount = viewport.Height * 0.10f;

            // Next, apply the amounts to shrink (yes, shrink) the viewport to zoom in
            viewport.Inflate(zoomWidthAmount, zoomHeightAmount);

            // And repaint
            Invalidate();
        }

        public void PanUp()
        {
            // First, get the height of the viewport, then calculate 10% of it
            float zoomHeightAmount = -viewport.Height * 0.10f;

            // Shift the viewport by this amount
            viewport.Offset(0.0f, zoomHeightAmount);

            // Finally, redraw the form
            Invalidate();
        }

        public void PanDown()
        {
            // First, get the height of the viewport, then calculate 10% of it
            float zoomHeightAmount = viewport.Height * 0.10f;

            // Shift the viewport by this amount
            viewport.Offset(0.0f, zoomHeightAmount);

            // Finally, redraw the form
            Invalidate();
        }

        // Reads longitude and latitude values from a text file
        private void LoadData()
        {
            /* This example stores latitude and longitude coordinates in the file Nebraska.txt.
             * Read in each line and turn it into a PointF object.
             */
            StreamReader reader = File.OpenText(@"..\..\Nebraska.txt");

            // Make a list to store the coordinates
            List<PointF> coordinates = new List<PointF>();

            // Read everything in
            while (!reader.EndOfStream)
            {
                /* Each line is in the form "latitude, longitude".  Parse each into a float, then
                 * create a PointF object from it.
                 */
                string[] values = reader.ReadLine().Split(',');

                // Parse both values into a PointF object
                coordinates.Add(
                    new PointF(
                    // The X coordinate (longitude) is the second word
                        float.Parse(values[1], CultureInfo.InvariantCulture),
                    // The Y coordinate (longitude) is the first word
                        float.Parse(values[0], CultureInfo.InvariantCulture)));
            }

            // Close the reader
            reader.Close();

            // Finally, make an array of projected coordinates
            geographicCoordinates = coordinates.ToArray();
        }

        protected override void OnKeyDown(KeyEventArgs e)
        {
            switch (e.KeyValue)
            {
                case 38:
                    PanUp();
                    break;
                case 40:
                    PanDown();
                    break;
            }
        }

        protected override void OnMouseWheel(MouseEventArgs e)
        {
            if (e.Delta < 0)
                ZoomOut();
            else
                ZoomIn();
        }

        protected override void OnMouseMove(MouseEventArgs e)
        {
            /* When the mouse is moved over the map, show the coordinate the mouse
             * is hovering over, in all three coordinate systems: pixel, projected, and geographic.
             */

            /* We use a Matrix to convert from projected coordinates to pixel coordinates.
             * The *inverse* of this matrix is used to convert pixels back into projected coordinates.
             */

            // First, make a copy of the current transform
            Matrix reverseTransform = transform.Clone();

            // Next, invert the transform
            reverseTransform.Invert();

            // We can now use this to calculate a projected coordinate.
            Point[] projectedCoordinate = new Point[] { e.Location };
            reverseTransform.TransformPoints(projectedCoordinate);

            /* "Points" now contains projected coordinates.  Use our projection to convert
             * this into geographic (latitude/longitude) coordinates.
             */

            PointF geographicCoordinate = plateCaree.Deproject(projectedCoordinate[0]);

            // Finally, display all three coordinate values: pixel, projected, geographic
            pixelCoordinateLabel.Text = "Pixel: " + e.Location.ToString();
            projectedCoordinateLabel.Text = "Projected: " + projectedCoordinate[0].ToString();
            geographicCoordinateLabel.Text = "Geographic: " + geographicCoordinate.ToString();
        }

        protected override void OnClosed(EventArgs e)
        {
            base.OnClosed(e);

            /* Clean up our unmanaged resources. 
             * These will get disposed of when the app ends, but this would be essential
             * if this were not the only form.
             */
            this.transform.Dispose();
            this.brush.Dispose();
            this.pen.Dispose();
        }
    }
}